
# THIS FILE IS GENERATED FROM NUMPY SETUP.PY
#
# To compare versions robustly, use `numpy.lib.NumpyVersion`
short_version = '1.14.0'
version = '1.14.0'
full_version = '1.14.0'
git_revision = 'a49ba685971392cbd9a9ca18d6cb17172dcae8ef'
release = True

if not release:
    version = full_version
